#include "ALMainWindow.h"
#include "util.h"
#include <QSettings>
#include <QTime>
#include <QtWidgets/qfiledialog.h>
#include <QMessageBox>
#include <stdlib.h>
#include <MatrixANN.h>

ALMainWindow::ALMainWindow(QWidget *parent) : QMainWindow(parent)
{

	setupUi(this);

	qsrand(QTime::currentTime().msecsSinceStartOfDay());

	loadWindowState();

	m_world = new World(this);
	m_world->SetContactListener(&m_contactListener);
	m_worldMutex = new QMutex();
	m_worldEngine = new WorldEngine(m_world, m_worldMutex, this);

	m_statisticsUpdateTimer.setInterval(500);
	m_statisticsUpdateTimer.start();

	m_worldTime = 0.0; // the start of time

	glwidget->setWorld(m_world, m_worldMutex);

	// statistics update all the time while the program is running
	connect(&m_statisticsUpdateTimer, SIGNAL(timeout()), this, SLOT(updateStatistics()));
	on_createWorldButton_clicked();
	on_createHerbivoresButton_clicked();
	on_createPlantsButton_clicked();
	on_createPredatorsButton_clicked();
	on_animateButton_clicked();
}



void ALMainWindow::on_createWorldButton_clicked()
{
	m_world->createBorder(b2Vec2(worldWidthSpinBox->value(), worldHeightSpinBox->value()));
}



void ALMainWindow::on_clearInhabitantsButton_clicked()
{
	m_world->clearInhabitants();
}


void ALMainWindow::on_createPlantsButton_clicked()
{
	int count = plantsCountSpinBox->value();

	CreatureConfiguration configuration;
	configuration.growthRate = plantsGrowthRateSpinBox->value();
	configuration.mutationRate = plantsMutationRateSizeSpinBox->value();
	double randomization = plantsMutationRateSizeSpinBox->value() * plantSizeSpinBox->value();
	configuration.energy = plantsEnergySpinBox->value();

	for (int i = 0; i < count; i++)
	{
		float x = qrand() / (float)RAND_MAX * m_world->size().x;
		float y = qrand() / (float)RAND_MAX * m_world->size().y;
			configuration.position = b2Vec2(x, y);
		configuration.size = Util::mutate(plantSizeSpinBox->value(), randomization);
		configuration.maxAge = Util::mutate(plantsMaxAgeSpinBox->value(), plantsMutationRateSizeSpinBox->value() * plantsMaxAgeSpinBox->value());
		m_world->addPlant(configuration);
	}
}

void ALMainWindow::on_createHerbivoresButton_clicked()
{
	int count = herbivoresCountSpinBox->value();

	CreatureConfiguration configuration;
	configuration.growthRate = herbivoresGrowthRateSpinBox->value();
	configuration.mutationRate = herbivoresMutationRateSizeSpinBox->value();
	double randomization = herbivoresMutationRateSizeSpinBox->value() * herbivoreSizeSpinBox->value();
	configuration.energy = herbivoresEnergySpinBox->value();

	for (int i = 0; i < count; i++)
	{
		float x = qrand() / (float)RAND_MAX * m_world->size().x;
		float y = qrand() / (float)RAND_MAX * m_world->size().y;
		configuration.position = b2Vec2(x, y);
		configuration.size = Util::mutate(herbivoreSizeSpinBox->value(), randomization);
		configuration.maxAge = Util::mutate(herbivoresMaxAgeSpinBox->value(), herbivoresMutationRateSizeSpinBox->value() * herbivoresMaxAgeSpinBox->value());
		m_world->addHerbivore(configuration);
	}
}

void ALMainWindow::on_createPredatorsButton_clicked()
{
	int count = predatorsCountSpinBox->value();

	CreatureConfiguration configuration;
	configuration.growthRate = predatorsGrowthRateSpinBox->value();
	configuration.mutationRate = predatorsMutationRateSizeSpinBox->value();
	double randomization = predatorsMutationRateSizeSpinBox->value() * predatorSizeSpinBox->value();
	configuration.energy = predatorsEnergySpinBox->value();

	for (int i = 0; i < count; i++)
	{
		float x = qrand() / (float)RAND_MAX * m_world->size().x;
		float y = qrand() / (float)RAND_MAX * m_world->size().y;
		configuration.position = b2Vec2(x, y);
		configuration.size = Util::mutate(predatorSizeSpinBox->value(), randomization);
		configuration.maxAge = Util::mutate(predatorsMaxAgeSpinBox->value(), predatorsMutationRateSizeSpinBox->value() * predatorsMaxAgeSpinBox->value());
		m_world->addPredator(configuration);
	}
}




void ALMainWindow::on_animateButton_clicked()
{
	if (glwidget->isAnimated())
	{
		glwidget->animate(false);
		animateButton->setText(tr("Start &Simulation"));
	}
	else
	{
		glwidget->animate(true);
		animateButton->setText(tr("Stop &Simulation"));
	}
}

void ALMainWindow::on_diaplayWorldCheckBox_toggled(bool checked)
{
	glwidget->enableWorldRender(checked);
}


void ALMainWindow::on_resetTimeButton_clicked()
{
	m_worldTime = 0.0;
}


void ALMainWindow::on_action_Save_triggered()
{
	QString filename = QFileDialog::getSaveFileName(this, tr("Save World As"));
	m_world->save(filename);
}


void ALMainWindow::on_browseStatisticsFileButton_clicked()
{
	QString filename = QFileDialog::getSaveFileName(this, tr("Select a file to save statistics in"));
	if (!filename.isEmpty())
	{
		statisticsFileLineEdit->setText(filename);
	}
	openStatisticsFileButton->click();
}

void ALMainWindow::on_openStatisticsFileButton_clicked()
{
	if (m_statisticsFile.isOpen())
	{
		m_statisticsFile.close();
		openStatisticsFileButton->setText(tr("&Open"));
		statisticsFileLineEdit->setEnabled(true);
	}
	else
	{
		m_statisticsFile.setFileName(statisticsFileLineEdit->text());
		bool opened = m_statisticsFile.open(QIODevice::WriteOnly);
		if (!opened)
		{
			QMessageBox::warning(this, tr("Cannot open file for writing"), tr("Please check permissions for %1").arg(m_statisticsFile.fileName()));
			return;
		}
		openStatisticsFileButton->setText(tr("&Close"));
		statisticsFileLineEdit->setEnabled(false);
	}
}


void ALMainWindow::updateStatistics()
{
	int plantsCount = 0;
	double plantsMinSize = 99999999;
	double plantsMaxSize = 0;
	double plantsSizeSum = 0;
	double plantsEnergySum = 0;

	int herbivoresCount = 0;
	double herbivoresMinSize = 99999999;
	double herbivoresMaxSize = 0;
	double herbivoresSizeSum = 0;
	double herbivoresEnergySum = 0;

	int predatorsCount = 0;
	double predatorsMinSize = 99999999;
	double predatorsMaxSize = 0;
	double predatorsSizeSum = 0;
	double predatorsEnergySum = 0;

	foreach(Creature *creature, m_world->population())
	{
		switch (creature->type())
		{
		default:
		case Creature::PLANT:
			plantsCount++;
			if (plantsMinSize > creature->size()) plantsMinSize = creature->size();
			if (plantsMaxSize < creature->size()) plantsMaxSize = creature->size();
			plantsSizeSum += creature->size();
			plantsEnergySum += creature->energy();
			break;

		case Creature::HERBIVORE:
			herbivoresCount++;
			if (herbivoresMinSize > creature->size()) herbivoresMinSize = creature->size();
			if (herbivoresMaxSize < creature->size()) herbivoresMaxSize = creature->size();
			herbivoresSizeSum += creature->size();
			herbivoresEnergySum += creature->energy();
			break;
		case Creature::HERBIVOREANN:
			herbivoresCount++;
			if (herbivoresMinSize > creature->size()) herbivoresMinSize = creature->size();
			if (herbivoresMaxSize < creature->size()) herbivoresMaxSize = creature->size();
			herbivoresSizeSum += creature->size();
			herbivoresEnergySum += creature->energy();
			break;

		case Creature::PREDATOR:
			predatorsCount++;
			if (predatorsMinSize > creature->size()) predatorsMinSize = creature->size();
			if (predatorsMaxSize < creature->size()) predatorsMaxSize = creature->size();
			predatorsSizeSum += creature->size();
			predatorsEnergySum += creature->energy();
			break;
		case Creature::PREDATORANN:
			predatorsCount++;
			if (predatorsMinSize > creature->size()) predatorsMinSize = creature->size();
			if (predatorsMaxSize < creature->size()) predatorsMaxSize = creature->size();
			predatorsSizeSum += creature->size();
			predatorsEnergySum += creature->energy();
			break;
		}
	}

	plantsCountLabel->setText(QString("%1").arg(plantsCount));
	plantsMinSizeLabel->setText(QString("%1").arg(plantsMinSize));
	plantsMaxSizeLabel->setText(QString("%1").arg(plantsMaxSize));
	plantsAvgSizeLabel->setText(QString("%1").arg(plantsSizeSum / plantsCount));
	plantsAvgEnergyLabel->setText(QString("%1").arg(plantsEnergySum / plantsCount));

	herbivoresCountLabel->setText(QString("%1").arg(herbivoresCount));
	herbivoresMinSizeLabel->setText(QString("%1").arg(herbivoresMinSize));
	herbivoresMaxSizeLabel->setText(QString("%1").arg(herbivoresMaxSize));
	herbivoresAvgSizeLabel->setText(QString("%1").arg(herbivoresSizeSum / herbivoresCount));
	herbivoresAvgEnergyLabel->setText(QString("%1").arg(herbivoresEnergySum / herbivoresCount));

	predatorsCountLabel->setText(QString("%1").arg(predatorsCount));
	predatorsMinSizeLabel->setText(QString("%1").arg(predatorsMinSize));
	predatorsMaxSizeLabel->setText(QString("%1").arg(predatorsMaxSize));
	predatorsAvgSizeLabel->setText(QString("%1").arg(predatorsSizeSum / predatorsCount));
	predatorsAvgEnergyLabel->setText(QString("%1").arg(predatorsEnergySum / predatorsCount));

	// if the world is running, update time
	if (glwidget->isAnimated())
	{
		m_worldTime += 11; // 0.5 s each update
	}

	if (m_statisticsFile.isOpen() && m_statisticsFile.isWritable() && glwidget->isAnimated())
	{
		QString line;
		line.append(QString("%1 ").arg(m_worldTime)); // world time
		line.append(QString("%1 %2 %3 %4 %5 ")
			.arg(plantsCount).arg(plantsMinSize).arg(plantsMaxSize).arg(plantsSizeSum / plantsCount).arg(plantsEnergySum / plantsCount));
		line.append(QString("%1 %2 %3 %4 %5 ")
			.arg(herbivoresCount).arg(herbivoresMinSize).arg(herbivoresMaxSize).arg(herbivoresSizeSum / herbivoresCount).arg(herbivoresEnergySum / herbivoresCount));
		line.append(QString("%1 %2 %3 %4 %5\n")
			.arg(predatorsCount).arg(predatorsMinSize).arg(predatorsMaxSize).arg(predatorsSizeSum / predatorsCount).arg(predatorsEnergySum / predatorsCount));

		m_statisticsFile.write(line.toUtf8());
		m_statisticsFile.flush();
	}
}


void ALMainWindow::loadWindowState()
{
	QSettings settings("settings.ini", QSettings::IniFormat);
	settings.beginGroup("WindowState");
	resize(settings.value("size", size()).toSize());
	move(settings.value("position", pos()).toPoint());
	action_World->setChecked(settings.value("viewWorld", action_World->isChecked()).toBool());
	action_Inhabitants->setChecked(settings.value("viewInhabitants", action_Inhabitants->isChecked()).toBool());
	action_Controls->setChecked(settings.value("viewControls", action_Controls->isChecked()).toBool());
	splitter->restoreState(settings.value("splitter").toByteArray());
	settings.endGroup();
}


void ALMainWindow::saveWindowState()
{
	QSettings settings("settings.ini", QSettings::IniFormat);
	settings.beginGroup("WindowState");
	settings.setValue("size", size());
	settings.setValue("position", pos());
	settings.setValue("viewWorld", action_World->isChecked());
	settings.setValue("viewInhabitants", action_Inhabitants->isChecked());
	settings.setValue("viewControls", action_Controls->isChecked());
	settings.setValue("splitter", splitter->saveState());
	settings.endGroup();
}




ALMainWindow::~ALMainWindow()
{
	saveWindowState();
	delete m_worldMutex;
}
