#include "creature.h"
#include "plant.h"
#include "util.h"
#include "herbivore.h"
#include "predator.h"
#include "predatorANN.h"
#include "herbivoreANN.h"
#include "world.h"
#include <QVector>
#include <QApplication>
#include <QtDebug>

Creature::Creature(World* world, const CreatureConfiguration& configuration, World* parent) : QObject(parent)
{
	this->world = world;
	m_box2DWorld = world->m_box2DWorld;
	m_initialSize = configuration.size;
	m_growthRate = configuration.growthRate;
	m_mutationRate = configuration.mutationRate;
	m_maxAge = configuration.maxAge;
	m_energy = configuration.energy;
	m_initialEnergy = configuration.energy;
	m_age = 0;
	m_reproductionTime = 0;
	m_paralyzationTimer = 0;
	m_reproductionTimeThreshold = 0.3 * m_maxAge;
}

Creature::Creature(const b2Vec2&, float size, World* world, World* parent) : QObject(parent)
{
	this->world = world;
	m_box2DWorld = world->m_box2DWorld;
	m_growthRate = 1.0001f;
	m_age = 0;
	m_maxAge = 120;
	m_initialSize = size;
	m_energy = 100;
	m_initialEnergy = m_energy;
	m_reproductionTime = 0;
	m_paralyzationTimer = 0;
	m_reproductionTimeThreshold = 0.3 * m_maxAge;
	m_mutationRate = 0.1f;
}


Creature::Type Creature::type()
{
	return DEFAULT;
}

void Creature::Step()
{
	grow();

	m_age += 1.0f / 60.0f;
	m_reproductionTime += 1.0f / 60.0f;

	if (isParalyzed()) m_paralyzationTimer -= 1.0f / 60.0f;
}

bool Creature::isDead()
{
	return m_age >= m_maxAge
		|| m_energy < 1;
}

bool Creature::canReproduce()
{
	return m_reproductionTime > m_reproductionTimeThreshold
		&& energyPercentage() > 0.60
		&& !isParalyzed();
}

float Creature::energyPercentage()
{
	return m_energy / m_initialEnergy;
}

float Creature::energy()
{
	return m_energy;
}

float Creature::size()
{
	return m_initialSize;
}

float Creature::mass()
{
	return m_body->GetMass();
}


void Creature::createSensor(float size)
{
	const float DEG2RAD = 3.14159f / 180.0f;
	QVector<b2Vec2> vertices;
	for (int i = 0; i < 360; i += 45)
	{
		float degInRad = i*DEG2RAD;
		vertices.append(b2Vec2(cos(degInRad)*size, sin(degInRad)*size));
	}
	b2PolygonShape shape;
	shape.Set(vertices.data(), vertices.length());

	b2FixtureDef fixturedef;
	fixturedef.shape = &shape;
	fixturedef.density = 0;
	fixturedef.isSensor = true;
	m_body->CreateFixture(&fixturedef);
}


QString Creature::toString()
{
	QString retVal;

	// type
	switch (type())
	{
	default:
	case PLANT:
		retVal.append("Plant\n");
		break;

	case HERBIVORE:
		retVal.append("Herbivore\n");
		break;

	case PREDATOR:
		retVal.append("Predator\n");
		break;
	}

	// position
	retVal.append(QString("%1 %2\n").arg(position().x).arg(position().y));

	// orientation
	retVal.append(QString("%1\n").arg(m_body->GetAngle()));

	// paralyzationTimer
	retVal.append(QString("%1\n").arg(m_paralyzationTimer));

	// energy
	retVal.append(QString("%1\n").arg(m_energy));

	// growthRate
	retVal.append(QString("%1\n").arg(m_growthRate));

	// age
	retVal.append(QString("%1\n").arg(m_age));

	// maxAge
	retVal.append(QString("%1\n").arg(m_maxAge));

	// mutationRate
	retVal.append(QString("%1\n").arg(m_mutationRate));

	// reproductionTime
	retVal.append(QString("%1\n").arg(m_reproductionTime));

	// reproductionTimeThreshold
	retVal.append(QString("%1\n").arg(m_reproductionTimeThreshold));

	// initialSize
	retVal.append(QString("%1\n").arg(m_initialSize));


	return retVal;
}



b2Body* Creature::body()
{
	return m_body;
}


b2Vec2 Creature::forceDirection()
{
	float angle = m_body->GetAngle();

	b2Vec2 retVal;
	retVal.x = cos(angle);
	retVal.y = sin(angle);
	return retVal;

}

void Creature::paralyze(float time)
{
	m_paralyzationTimer = time;
}

bool Creature::isParalyzed()
{
	return m_paralyzationTimer > 0;
}


void Creature::go(float force)
{
	if (isParalyzed()) return;
	m_body->ApplyForce(force * forceDirection(), m_body->GetWorldCenter(), true);
	// E = (mv^2)/2
	// v = at
	// F = ma
	// E = F^2 t^2 / (2m)
	m_energy -= force * force / 3600 / 2 / m_body->GetMass();
}

void Creature::turn(float torque)
{
	if (isParalyzed()) return;
	m_body->ApplyTorque(torque, true);
}

void Creature::turnTo(float angle)
{
	if (isParalyzed()) return;

	float nextAngle = m_body->GetAngle() + m_body->GetAngularVelocity() / 15.0;
	float totalRotation = angle - nextAngle;
	while (totalRotation < DEGTORAD(-180)) totalRotation += DEGTORAD(360);
	while (totalRotation > DEGTORAD(180)) totalRotation -= DEGTORAD(360);
	float desiredAngularVelocity = totalRotation * 10;
	float torque = m_body->GetInertia() * desiredAngularVelocity / (1 / 60.0);

	turn(torque);
}


void Creature::explore()
{
	float torque = 50 * m_body->GetMass();

	int r = (qrand() / (float)RAND_MAX) * 10; // 1 in 10
	if (r == 5) // 5 is just good as any
	{
		turn((qrand() / (float)RAND_MAX) > 0.5 ? torque : -torque);
		go(200);
	}
}



b2Vec2 Creature::position()
{
	return m_body->GetPosition();
}


void mix_ANN_weigths(ANN* p1, ANN* p2,ANN* child) {
	//arithmetic crossing beetween wieigths p1 and p2 and mutation{mutation types,random add,swap of weigths}
	auto layers = child->weights;
	for (int i = 0; i < layers.size(); i++){
		for (int j = 0; j < layers[i].size(); j++){
			for (int k = 0; k < layers[i][j].size(); k++){
				layers[i][j][k] = (p1->weights[i][j][k] + p2->weights[i][j][k]) / 2.0f + p1->randomize_gauss();
			}
		}
	}
}

Creature* Creature::reproduce(Creature* other)
{
	// new initial values
	b2Vec2 pos = Util::mutate(0.5 * (m_body->GetPosition() + other->m_body->GetPosition()), m_initialSize * m_mutationRate);
	double size = (m_initialSize + other->m_initialSize) / 2;
	size = Util::mutate(size, size * m_mutationRate);
	if (size < 0.01) size = 0.01;

	// create new creature
	Creature *newCreature;
	switch (type())
	{
	default:
	case PLANT:
		newCreature = new Plant(pos, size, world, (World*)parent());
		break;

	case HERBIVORE:
		newCreature = new Herbivore(pos, size, world, (World*)parent());
		break;

	case PREDATORANN:
		newCreature = new PredatorANN(pos, size, world, (World*)parent());
		mix_ANN_weigths(((PredatorANN*)this)->neural, ((PredatorANN*)other)->neural, ((PredatorANN*)newCreature)->neural);
		break;
	case HERBIVOREANN:
		newCreature = new HerbivoreANN(pos, size, world, (World*)parent());
		mix_ANN_weigths(((HerbivoreANN*)this)->neural, ((HerbivoreANN*)other)->neural, ((HerbivoreANN*)newCreature)->neural);
		break;
	}

	// mix between parents + mutate
	double tmp = (m_growthRate + other->m_growthRate) / 2;
	newCreature->m_growthRate = Util::mutate(tmp, tmp * pow(m_mutationRate, 10));

	tmp = (m_maxAge + other->m_maxAge) / 2;
	newCreature->m_maxAge = Util::mutate(tmp, tmp * m_mutationRate);

	tmp = (m_reproductionTimeThreshold + other->m_reproductionTimeThreshold) / 2;
	newCreature->m_reproductionTimeThreshold = Util::mutate(tmp, tmp * m_mutationRate);

	float thisEnergy = m_energy > m_initialEnergy ? m_initialEnergy : m_energy;
	float otherEnergy = other->m_energy > other->m_initialEnergy ? other->m_initialEnergy : other->m_energy;
	tmp = (thisEnergy + otherEnergy) / 2;
	newCreature->m_energy = Util::mutate(tmp, tmp * m_mutationRate);
	newCreature->m_initialEnergy = newCreature->m_energy;

	// reset reproduction
	m_reproductionTime = 0;
	//     other->m_reproductionTime = 0;

	((World*)parent())->addCreature(newCreature);

	return newCreature;
}

void Creature::addToNearBy(Creature* creature)
{
	// predators in front, everything else whatever
	if (creature->type() == PREDATOR)
		m_nearByCreatures.prepend(creature);
	else
		m_nearByCreatures.append(creature);
}


void Creature::addToTouching(Creature* creature)
{
	// predators in front, everything else whatever
	if (creature->type() == PREDATOR)
		m_touchingCreatures.prepend(creature);
	else
		m_touchingCreatures.append(creature);
}

void Creature::removeFromNearBy(Creature* creature)
{
	m_nearByCreatures.removeOne(creature);
}

void Creature::removeFromTouching(Creature* creature)
{
	m_touchingCreatures.removeOne(creature);
}

Creature* Creature::getClosestVisible(Creature::Type type)
{
	Creature *retVal = 0;

	//radijus kruga u euklidskoj udaljenosti koji je zapravo kvadrat
	int r = 0;

	//Kvadrant u kojem se nalazi stvorenje
	int init_x = this->position().x / world->m_size.x * world->wh;
	int init_y = this->position().y / world->m_size.y * world->wh;

	//qDebug() << init_x << "," << init_y << endl;

	while (!retVal && r < world->wh || r == 1) {
		//		qDebug() << this->type() << " " << type << endl;
		//		qDebug() << r << endl;
		//		qDebug() << init_x << " " <<init_y << endl;
		//pretraga po okomicama odozgo
		for (int x = max(init_x - r, 0); x < min(world->wh - 1, init_x + r); x++)
		{
			if (init_y - r >= 0)
			{
				retVal = getClosestVisible(searchQuadrant(x, init_y - r, type), retVal);
			}

			if (init_y + r < world->wh)
			{
				retVal = getClosestVisible(searchQuadrant(x, init_y + r, type), retVal);
			}
		}

		for (int y = max(0, init_y - r + 1); y < min(world->wh - 1, init_y + r) - 1; y++)
		{
			if (init_x - r > 0)
			{
				retVal = getClosestVisible(searchQuadrant(init_x - r, y, type), retVal);
			}

			if (init_x + r < world->wh)
			{
				retVal = getClosestVisible(searchQuadrant(init_x + r, y, type), retVal);
			}
		}
		//		system("pause");
		r++;
	}
	return retVal;
}

Creature* Creature::getClosestVisible(Creature* tmp, Creature* retVal)
{
	if (tmp)
	{
		if (!retVal) {
			retVal = tmp;
		}
		else
		{
			if (Util::distance(this->position(), tmp->position()) < Util::distance(this->position(), retVal->position()))
			{
				retVal = tmp;
			}
		}
	}
	return retVal;
}

Creature* Creature::searchQuadrant(int x, int y, Type type)
{
	Creature* retVal = 0;
	float minDistance = -1;

	foreach(Creature *creature, world->grid[x][y])
	{
		if (creature->type() == type)
		{
			double d = Util::distance(creature->position(), position());
			if (minDistance < 0 || minDistance > d)
			{
				minDistance = d;
				retVal = creature;
			}
		}
	}

	return retVal;
}

void Creature::grow(float growthRate)
{
	if (growthRate == 0)
	{
		growthRate = m_growthRate;
		growthRate = 1 + (growthRate - 1) / ((m_age > 1) ? sqrt(m_age) / 10 : 1); // reduce the decimal side of the growthRate
	}

	float oldMass = m_body->GetMass();

	if (growthRate < 1 && oldMass < 1) return;

	b2Fixture *fixture = m_body->GetFixtureList();

	do {
		b2PolygonShape* shape = (b2PolygonShape*)fixture->GetShape();

		for (int i = 0; i < 8; i++)
		{
			shape->m_vertices[i].x *= growthRate;
			shape->m_vertices[i].y *= growthRate;
		}
	} while ((fixture = fixture->GetNext()));

	m_body->ResetMassData();

	m_energy -= 10 * (m_body->GetMass() - oldMass);
}

void Creature::shrink(float shrinkRate)
{
	if (shrinkRate == 0) shrinkRate = pow(2 - m_growthRate, 2);
	grow(shrinkRate);
}

float Creature::takeEnergy(Creature *taker)
{
	b2Vec2 v = taker->m_body->GetLinearVelocity();
	float speed = sqrt(v.x*v.x + v.y*v.y);
	float e = taker->m_body->GetMass() * speed + 10; // static part +10
	m_energy -= e;
	return e;
}


Creature::~Creature()
{
}
