
#include <QtWidgets>
#include <QtOpenGL>
#include <QOpenGLFunctions>
#include <math.h>

#include "glwidget.h"
#include "plant.h" // testing only

#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

GLWidget::GLWidget(QWidget* parent)
    : QGLWidget(QGLFormat(QGL::SampleBuffers), parent)
{
    m_updateTimer = new QTimer(this);
    connect(m_updateTimer, SIGNAL(timeout()), this, SLOT(updateGL()));
    m_updateTimer->start(1000/60);
    
    m_pan.x = m_pan.y = 0;
    m_zoomLevel = 1;
    m_animationEnabled = false;
    m_worldRenderingEnabled = true;
}

GLWidget::~GLWidget()
{
}


void GLWidget::setWorld(World* world, QMutex* worldMutex)
{
    m_world = world;
    m_worldMutex = worldMutex;
}


b2Vec2 GLWidget::screenToWorld(const QPointF& screenCoordinate)
{
    b2Vec2 retVal;
    retVal.x = screenCoordinate.x() * m_zoomLevel - m_pan.x;
    retVal.y = screenCoordinate.y() * m_zoomLevel - m_pan.y;
    
    return retVal;
}


void GLWidget::drawCreature(Creature* creature)
{
    b2Body *body = creature->body();
    b2Vec2 pos = body->GetWorldCenter();
    
    b2Fixture *fixture = body->GetFixtureList();
    do {
	if( fixture->IsSensor() ) continue; // don't draw sensors
	
	glPushMatrix();
	    glTranslatef( pos.x, pos.y, 0);
	    glRotatef(body->GetAngle()*180.0/M_PI,0,0,1);
	
	    float transparency = creature->energyPercentage();
	    // inner color
	    switch( creature->type() )
	    {
		default:
		case Creature::DEFAULT:
		    glColor4f(0.95, 0.95, 0.95, transparency);
		    break;
		    
		case Creature::PLANT:
		    glColor4f(0.3, 0.95, 0.3, transparency);
		    break;
		    
		case Creature::HERBIVORE:
		    glColor4f(0.3, 0.3, 0.95, transparency);
		    break;
		    
		case Creature::PREDATOR:
		    glColor4f(0.95, 0.3, 0.3, transparency);
		    break;
		case Creature::HERBIVOREANN:
			glColor4f(0.3, 0.3, 0.95, transparency);
			break;

		case Creature::PREDATORANN:
			glColor4f(0.95, 0.3, 0.3, transparency);
			break;
	    }
	    
	    
	    // draw solid area
	    glBegin(GL_TRIANGLE_FAN);
		b2PolygonShape* shape = (b2PolygonShape*)fixture->GetShape();
		int vertexCount = shape->GetVertexCount();
		for( int i = 0; i < vertexCount; i++ )
		{
		    glVertex2f( shape->GetVertex(i).x, shape->GetVertex(i).y);
		}
	    glEnd();
	    
	    // draw head (first point)
	    if( creature->type() != Creature::PLANT )
	    {
		glColor3f(1, 0, 1);
		glPointSize(2);
		glBegin(GL_POINTS);
		    glVertex2f( shape->GetVertex(0).x, shape->GetVertex(0).y );
		glEnd();
	    }
	
	    // outline color
	    switch( creature->type() )
	    {
		default:
		case Creature::DEFAULT:
		    glColor3f(1, 1, 1);
		    break;
		    
		case Creature::PLANT:
		    glColor3f(0, 1, 0);
		    break;
		    
		case Creature::HERBIVORE:
		    glColor3f(0, 0, 1);
		    break;
		    
		case Creature::PREDATOR:
		    glColor3f(1, 0, 0);
		    break;
		case Creature::HERBIVOREANN:
			glColor3f(0, 0, 1);
			break;

		case Creature::PREDATORANN:
			glColor3f(1, 0, 0);
			break;
	    }
	    
	    if( creature->isParalyzed() )
		glColor3f( 0.5, 0, 1 );
	    
	    // draw outline
	    glLineWidth(2); //fat lines
	    glBegin(GL_LINE_LOOP);
		for( int i = 0; i < vertexCount; i++ )
		{
		    glVertex2f( shape->GetVertex(i).x, shape->GetVertex(i).y);
		}
	    glEnd();
	    
	glPopMatrix();
    } while( (fixture = fixture->GetNext()) );
}


void GLWidget::drawB2Body(const b2Body* body)
{
    b2Vec2 pos = body->GetWorldCenter();
    
    glPushMatrix();
	glColor3f( 128, 128, 128 );
        glTranslatef( pos.x, pos.y, 0);
        glRotatef(body->GetAngle()*180.0/M_PI,0,0,1);
        glBegin(GL_TRIANGLE_FAN);
            b2PolygonShape* shape = (b2PolygonShape*)body->GetFixtureList()->GetShape();
            int vertexCount = shape->GetVertexCount();
            for( int i = 0; i < vertexCount; i++ )
            {
                glVertex2f( shape->GetVertex(i).x, shape->GetVertex(i).y);
            }
        glEnd();
    glPopMatrix();
}


void GLWidget::pan(const QPointF& amount)
{
    m_pan.x += (amount.x()) * m_zoomLevel;
    m_pan.y += (amount.y()) * m_zoomLevel;
}


void GLWidget::zoom(double factor)
{
    m_zoomLevel /= factor;
}

void GLWidget::animate(bool enable)
{
    m_animationEnabled = enable;
}

void GLWidget::enableWorldRender(bool enable)
{
    m_worldRenderingEnabled = enable;
}


bool GLWidget::isAnimated()
{
    return m_animationEnabled;
}


void GLWidget::initializeGL()
{
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    
    qglClearColor(QColor(Qt::black));

    glMatrixMode(GL_PROJECTION);
        glOrtho(-m_pan.x, m_zoomLevel * width() - m_pan.x,m_zoomLevel * height() - m_pan.y, -m_pan.y, -1, 1);
    glMatrixMode(GL_MODELVIEW);
    glClearColor(0,0,0,1);
}

void GLWidget::paintGL()
{
    if( m_worldRenderingEnabled )
    {
	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();
	
	// draw world
	foreach( b2Body *b, m_world->border() )
	{
	    drawB2Body(b);
	}
	
	// draw population
	foreach( Creature *creature, m_world->population() )
	{
	    drawCreature( creature );
	}
    }
    
    if( m_animationEnabled )
	m_world->Step(1.0/60.0,8,3);
}

void GLWidget::resizeGL(int width, int height)
{
    glViewport( 0, 0, width, height );

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
#ifdef QT_OPENGL_ES_1
    glOrthof(-m_pan.x, m_zoomLevel * width - m_pan.x,m_zoomLevel * height - m_pan.y, -m_pan.y, -1, 1);
#else
    glOrtho(-m_pan.x, m_zoomLevel * width - m_pan.x,m_zoomLevel * height - m_pan.y, -m_pan.y, -1, 1);
#endif
    glMatrixMode(GL_MODELVIEW);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    m_lastPos = event->pos();
    if( event->buttons() & Qt::RightButton )
    {
        int r = rand()/(float)RAND_MAX * 3;
        switch( r )
        {
            default:
            case 0:
                m_world->addPlant(screenToWorld(event->pos()), 3);
                break;
            case 1:
                m_world->addHerbivore(screenToWorld(event->pos()), 3);
                break;
            case 2:
                m_world->addPredator(screenToWorld(event->pos()), 3);
                break;
        }
    }
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    if( event->buttons() & Qt::LeftButton )
    {
        pan(event->pos() - m_lastPos);
        resizeGL(width(), height());
    }
    
    m_lastPos = event->pos();
}

void GLWidget::wheelEvent(QWheelEvent* event)
{
    QPointF center;
    center.setX( width() / 2 );
    center.setY( height() / 2 );
    
    if( event->angleDelta().y() < 0 )
    {
        zoom(0.5);
        pan(event->pos() * 0.5);
    }
    else
    {
        zoom(2);
        pan( -event->pos() );
    }
    
    
    
    resizeGL( width(), height() );
}

